#!/usr/bin/env python

import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from jmarkov.phase.dtph import dtph

alpha = np.array([0.05,0.9,0,0.05])
T = np.array([[0.2, 0.7, 0, 0.1], [0.3, 0.2, 0.2,0.1], [0.1, 0, 0.5,0.1],[0.2,0.2,0.2,0.2]])

ph = dtph(alpha, T)
print(ph.alpha)
print(ph.T)

pmf1 = ph.pmf(5)
print(pmf1)
pmf2 = ph.pmf(10)
print(pmf2)
ev = ph.expected_value()
print(ev)




#!/usr/bin/env python

import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from jmarkov.ctmc import ctmc


Q = np.array([[-4, 1, 3], [2, -5, 3], [1, 2, -3]])

mc = ctmc(Q)
print(mc.generator)

ss = mc.steady_state()
print(ss)
print("lets try the transient method:")
a=mc.transient_probabilities(1,np.array([0.1,0.5,0.4]))
print(a)

F=ctmc(np.array([[-3,2,1],[2,-5,3],[1,1,-2]]))
print("Lets try first passage time in ctcm with this matrix: ")
print(np.array([[-3,2,1],[2,-5,3],[1,1,-2]]))
print("This is the result: ")
print(F.first_passage_time(0))

m=ctmc(np.array([[-10, 10, 0, 0, 0, 0], [15, -25, 10, 0, 0, 0], [0, 15, -25, 10, 0, 0], 
                 [0, 0, 15, -25, 10, 0], [0, 0, 0, 15, -25, 10], [0, 0, 0, 0, 15, -15]]))
print(m.occupation_time(1))





import unittest
import sys
import os.path
import numpy as np
from numpy.testing import assert_allclose
# context for the tests
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from jmarkov.ctmc import ctmc


class TestSteadyState(unittest.TestCase):
    def test_ctmc_steady_state(self):
        Q = np.array([[-4, 1, 3], [2, -5, 3], [1, 2, -3]])
        mc = ctmc(Q)
        assert_allclose(mc.steady_state(), [0.25, 0.25, 0.5],err_msg="should be [0.25, 0.25, 0.5]")

class TestIsErgodic(unittest.TestCase):
    def test_ctmc_isergodic(self):
        Q = np.array([[-4, 1, 3], [2, -5, 3], [1, 2, -3]])
        mc = ctmc(Q)
        self.assertTrue(mc.is_ergodic())

if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python

import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from jmarkov.phase.ctph import ctph


alpha = np.array([1, 0, 0])
T = np.array([[-4, 1, 3], [2, -5, 2], [0, 2, -3]])

ph = ctph(alpha, T)
print(ph.alpha)
print(ph.T)

pdf1 = ph.pdf(0.5)
print(pdf1)

pdf2 = ph.pdf(5)
print(pdf2)

ev = ph.expected_value()
print(ev)



import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..','..','..')))
from jmarkov.dtmc import dtmc

estados = np.array(["Fr", "Al","In", "Es"])

matriz = np.array([[0.5, 0.2, 0, 0.3],
                   [0.15, 0.55, 0.15, 0.15],
                   [0.25, 0.25, 0.25, 0.25],
                   [0, 0.3, 0.7, 0]])

viajero = dtmc(matriz)


import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..','..','..')))
from jmarkov.mdp.dtmdp import dtmdp

# Variable de estado:
# X_{t}: Clasificación de Fontibón en la época t
# Y_{t}: Clasificación de Usme en la época t
# Z_{t}: (X_{t},Y_{t})

estadosFontibon = ['Normal', 'Alerta'] 
estadosUsme = ['Normal', 'Alerta'] 

estadosAcueducto = []
for i in estadosFontibon:
    for j in estadosUsme:
        estado = ",".join((str(i),str(j)))
        estadosAcueducto.append(estado)
estadosAcueducto = np.array(estadosAcueducto)

# Acciones
accionesAcueducto = np.array(['Nada','Fontibón','Usme'])
# Retornos Inmediatos
retornosAcueducto = np.array([[0, 90, 55],
                              [149, 239, 175],
                              [183, 180, 238],
                              [332, 329, 358]])
# Matrices de transición
matricesAcueducto = {}
# Para a == Nada
matricesAcueducto[accionesAcueducto[0]] = np.array([[0.09, 0.21, 0.21, 0.49],
                                   [0, 0.3, 0, 0.7],
                                   [0, 0, 0.3, 0.7],
                                   [0, 0, 0, 1]])  
# Para a == Fontibón
matricesAcueducto[accionesAcueducto[1]] = np.array([[0.21, 0.49, 0.09, 0.21],
                                   [0, 0.7, 0, 0.3],
                                   [0.24, 0.56, 0.06, 0.14],
                                   [0, 0.8, 0, 0.2]])
# Para a == Usme
matricesAcueducto[accionesAcueducto[2]] = np.array([[0.21, 0.09, 0.49, 0.21],
                                   [0.24, 0.06, 0.56, 0.14],
                                   [0, 0, 0.7, 0.3],
                                   [0, 0, 0.8, 0.2]])

# Creo el problema como un mdp
mdpAcueducto = dtmdp(estadosAcueducto, accionesAcueducto, matricesAcueducto, retornosAcueducto, 0.8)
print(mdpAcueducto.solve(0, minimize = True, method = "value_iteration"))
print(mdpAcueducto.solve(0, minimize = True, method="policy_iteration"))


import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..','..','..')))
from jmarkov.mdp.dtmdp import dtmdp

# Variable de estado:
# X_{t}: Incidencia de la moniliasis en los cultivos de cacao en la época t

estadosAcacias = np.array(['Baja', 'Media','Alta'])

# Acciones
accionesAcacias = np.array(['Erradicar (E)','Fumigar (F)','No hacer nada (N)'])
# Retornos Inmediatos
retornosAcacias = np.array([[-450, -125, -50],
                              [-800, -1000,-400],
                              [-1000, -1500, -600]])
# Matrices de transición
matricesAcacias = {}
# Para a == Nada
matricesAcacias['Erradicar (E)'] = np.array([[0.95, 0.05, 0],
                                 [0.8, 0.15, 0.05],
                                 [0, 0.8, 0.2]])  
# Para a == Fontibón
matricesAcacias['Fumigar (F)'] = np.array([[0.8, 0.2, 0],
                                 [0.5, 0.3, 0.2],
                                 [0, 0.5, 0.5]])
# Para a == Usme
matricesAcacias['No hacer nada (N)'] = np.array([[0.6, 0.4, 0],
                                 [0.1, 0.5, 0.4],
                                 [0, 0.1, 0.9]])

# Creo el problema como un mdp
mdpAcacias = dtmdp(estadosAcacias, accionesAcacias, matricesAcacias, retornosAcacias, 0.8)
print(mdpAcacias.solve(0, method = "value_iteration"))
print(mdpAcacias.solve(0, method="policy_iteration"))


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import sys
import os.path

# context for the autodoc
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'jMarkov'
copyright = '2024, COPA@Uniandes'
author = 'juanfperez'
release = '0.1'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.githubpages'
]

templates_path = ['_templates']
exclude_patterns = []



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output
#import sphinx_rtd_theme
html_theme = 'alabaster'
#html_theme = 'sphinx_rtd_theme'
#html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
html_static_path = ['_static']



import numpy as np
from scipy import linalg
from scipy import sparse
from .markov_chain import markov_chain
import math

class dtmc(markov_chain):
    """
    Implements a finite discrete-time Markov chain (CTMC) 
    
    The chain is defined by its number of states, states, and a 
    transition matrix. The class provides methods to compute both
    stationary and transient metrics, as well as to check the 
    chain properties (aperiodicity and ergodicity). 
    """

    # number of states in string array form
    n_states:int=1

    # states in string array form   
    states:np.array = np.array([1])

    # transition matrix as 2d numpy array
    transition_matrix:np.array
    
    # empty initializer 
    # chain with a single state
    def __init__(self):
        """
        Creates a continuous-time Markov chain from its transition matrix
        """
        self.n_states = 1
        self.transition_matrix = np.array([1])

    # initializer with only the number of states
    def __init__(self,n:int):
        self.n_states = n
        self.transition_matrix = np.eye(n, dtype=float)

    # initializer with a transition matrix
    def __init__(self,transition_matrix:np.array):
        if not self._check_transition_matrix(transition_matrix):#Lets check if transition matrix is logical (i.e the rows sum 1)
            raise ValueError("the rows of transition matrix do not sum 1 or has non positive values")
        self.n_states=transition_matrix.shape[0]
        self.transition_matrix = transition_matrix

    def _check_transition_matrix(self, M:np.array):
        """
        Checks that a matrix is a Markov chain transition matrix
         
        Checks that all entries are non-negative and all row sums are equal to one
        """
        #Check if a given transition matrix has the condition that for every row the sum of all elements is equal to 1
        #Check if a given transition matrix has the condition that every element of the matrix is between 0 and 1
        if np.all(np.isclose(np.sum(M,axis=1),1,1e-5)) and np.all((M>=0) & (M<=1)):
            return True
        else:
            return False
        

    def steady_state(self) -> np.ndarray:
        """
        Computes the steady state distribution of the discrete-time Markov chain

        Computes the steady state probability distribution by replacing one of the 
        matrix equations with a normalizing equation that ensures the result is a 
        probability distribution. 

        Returns the stationary probability distribution in array form
        """
        # computes the steady state distribution
        if self.transition_matrix.any():
            # sets A to be I-P and replace first column with ones for the normalizing equation
            A = self.transition_matrix - np.eye(self.n_states, dtype=float)
            A[:, 0] =  1
            # sets b as the right hand vector with a 1 for the normalizing equation
            b = np.zeros(self.n_states, dtype=float)
            b[0] = 1
            # solves transposed system as pi is a row vector
            res = linalg.solve(A,b,transposed=True)
            return res
        else:    
            print("Empty transition matrix")
        return 0
    

    def transient_probabilities(self,n,alpha):
        """
        Computes the transient distribution at transition n with initial state alpha

        Computes alpha*(P^n)*ones to obtain the probability distribution
        at time/transition n given the initial probability distribution alpha
        """
         
        #First, lets verify n is integer
        if not isinstance(n,int):
            raise ValueError("The number of transitions n must be integer")
        #Now lets verify the dimensions of the vector alpha:
        try:
            shape=alpha.shape
        except:
            print("alpha vector must be a numpy array")
            return None
        if shape[0]!=self.n_states:
            raise ValueError("The dimensions of alpha vector are incorrect. It must be a vector 1xn_states")
        
        #Computes transient_matrix**n
        matrix_n=np.linalg.matrix_power(self.transition_matrix,n)
        vector=alpha@matrix_n
        return vector


    def first_passage_time(self, target:str):
        """
        Computes the expected first passage time to a target state from a start state.

        This method calculates the expected number of steps required for the Markov chain to reach
        the specified target state from the start state by creating a sub-matrix of the transition matrix with the target removed.

        Returns the expected steps to reach the target state from any start state (except target) in array form        
        """
        #The transition matrix and the number of states are brought
        n=self.n_states 
        M=self.transition_matrix.copy()
        # The identity matrix is created with size n-1 states
        i=np.identity(n-1)
        # Matrix of ones with n-1 rows is created
        u=np.full([n-1,1],1)
        # The column and row corresponding to the target state are eliminated from the transition matrix
        p=np.delete(M,target, axis=0)
        p=np.delete(p,target, axis=1)
        
        t=np.matmul(np.linalg.inv(i-p),u)
        return t 

    def occupation_time(self, n:int):
        """
        Computes the expected occupation time in n steps.

        This method calculates the expected time that the discrete-time Markov chain will remain in each state, from all starting states.


        Returns the occupation matrix in array form  
        """
        # Create an identity matrix with the same shape of transition matrix
        occupation=np.eye(self.transition_matrix.shape[0],self.transition_matrix.shape[1])
        # Iterate for each step
        for i in range(1,n+1):
            # Calculate the occupation time for each step, and update the occupation matrix
            occupation+=np.linalg.matrix_power(self.transition_matrix,i)
        return occupation
    
    def period(self):
        """
        Computes the period of a discrete-time Markov Chain.
        
        This method calculates the period of a discrete-time Markov Chain by iterating a total of N steps, where N represents
        the number of states that the chain has. In each k-th iteration, the chain is elevated to the k-th power to compute the
        transient probabilities in the k-th step. If one of the elementsin in the diagonal of the resulting matrix is greater than 0,
        the chain can potentially have a period k. The period of the matrix will be the Greatest Common Divisor (GCD) between the
        potential new period of the matrix in the k-th iteration and the current GCD before the k-th iteration.

        Returns the period of the chain as an int
        """
        
        #Initializes the transition matrix and the greatest common divisor for the periodicity calculations
        P_k = self.transition_matrix
        gcd = 0
        
        #Loop to elevate the transition matrix to the k th power and evaluate its periodicity:
        for k in range(2, (self.n_states)+1):
            P_k = np.dot(self.transition_matrix, P_k)
            #If the transition matrix to the k-th power has a non-zero element in its diagonal, the chain could have a period k:
            if np.any(np.diagonal(P_k) != 0):
                #GCD between the current GCD for the period of the matrix and the potential new GCD
                gcd = math.gcd(gcd, k)
            #If the GCD is 1, there is no need to keep looking for the period, the chain is aperiodic:
            if gcd == 1:
                break
        return gcd
    
    def is_irreducible(self):
        """
        Checks if the given discrete-time Markov Chain is irreducible.
        
        This method determines if the discrete-time Markov Chain is irreducible by checking if, starting in
        any state, it is possible to reach any other state in a sequence of transitions.

        Returns a boolean
        """
        # Brings a copy of the transition matrix
        P = np.copy(self.transition_matrix)
        # Checks if all the states in the Markov chain are strongly connected
        if sparse.csgraph.connected_components(P, directed=True,connection='strong',return_labels=False)==1:
            return True
        else:
            return False

    def is_ergodic(self):
        """
        Checks if a given discrete-time Markov Chain is ergodic.
        
        Given that a finite discrete-time Markov Chain is ergodic if it is aperiodic and irreducible, this
        method checks if both conditions are met to determine if the provided chain is ergodic or not.
        
        Returns a boolean
        """
        # Checks if the chain is irreducible and if it has a period of 1
        if self.period() == 1 & self.is_irreducible() == True:
            return True
        else:
            return False
    






import sys
import os.path
import numpy as np
# context for the examples
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..','..','..')))
from jmarkov.mdp.dtmdp import dtmdp


# Parámetros:
q = 0.6
r = 0.4
discount_factor = 0.8
# Variables de estado:
# X_{t}: Estado de la batería en la época t
# Y_{t}: Horas acumuladas de la batería en la época t
# Z_{t}: (X_{t},Y_{t})
# Espacio de estados
Sx = [i for i in range(1,6)] # [1, 2, 3, 4, 5]
Sy = [j for j in range(0,100500,500)] # [0, 500,1000, ..., 99500, 100000]

estados = []
for i in Sx:
    for j in Sy:
        estado = ",".join((str(i),str(j)))
        estados.append(estado)
estados = np.array(estados)

# Acciones
acciones = np.array([a for a in range(0,3)]) # array([0, 1, 2])--> 0: no hacer nada. 1: reparar. 2: reemplazar.

# Retornos Inmediatos
retornos = []
for estado_actual in range(0,len(estados)):
    i = int(estados[estado_actual].split(",")[0])
    j = int(estados[estado_actual].split(",")[1])
    listaAux = []
    for a in acciones:
        if a == 0 and i<5 and j<100000:
            listaAux.append(0)
        elif a == 1 and i<5 and j<100000:
            listaAux.append(1)
        elif a == 2 and i==5:
            listaAux.append(100)
        elif a == 2 and j==100000:
            listaAux.append(100)
        elif a==0 and i==5:
            listaAux.append(1000)
        elif a==0 and j==100000:
            listaAux.append(1000)
        elif a==1 and i==5:
            listaAux.append(1000)
        elif a==1 and j==100000:
            listaAux.append(1000)
        elif a==2 and i<5:
            listaAux.append(1000)
        elif a==2 and j<100000:
            listaAux.append(1000)
    retornos.append(listaAux)
retornos = np.array(retornos)

# Matrices de transición
matrices = {}
for a in acciones:
    # Para a == 0 ~ no hacer nada
    if a == 0:
        matriz = []
        for estado_actual in range(0,len(estados)):
            i = int(estados[estado_actual].split(",")[0])
            j = int(estados[estado_actual].split(",")[1])
            listaAux = []
            for estado_futuro in range(0,len(estados)):
                i_ = int(estados[estado_futuro].split(",")[0])
                j_ = int(estados[estado_futuro].split(",")[1])
                # Que se mantenga el nivel y aumenten las horas en 500
                if i_ == i and j_==j+500 and i<5 and j<99500:
                    listaAux.append((1-q)*(1-r))
                # Que aumente el nivel y aumenten las horas en 500
                elif i_==i+1 and j_==j+500 and i<5 and j<99500:
                    listaAux.append(q*r)
                # Que se mantenga el nivel y aumenten las horas en 1000
                elif i_==i and j_==j+1000 and i<5 and j<99500:
                    listaAux.append((1-q)*r)
                # Que aumente el nivel y aumenten las horas en 1000
                elif i_==i+1 and j_==j+1000 and i<5 and j<99500:
                    listaAux.append(q*(1-r))
                # Que se mantenga el nivel y el aumento de horas me haga llegar a 100000 (casos extremos)
                elif i_==i and j_==100000 and i<5 and j>99000:
                    listaAux.append(((1-q)*r)+((1-q)*(1-r)))
                # Que aumente el nivel y el aumento de horas me haga llegar a 100000 (casos extremos)
                elif i_==i+1 and j_==100000 and i<5 and j>99000:
                    listaAux.append((q*(1-r))+(q*r))
                # Que aumente o se mantenga el nivel y aumenten las horas en 500 pero estoy en 5 así que el nivel no puede empeorar más
                elif i_==i and j_==j+500 and i==5 and j<99500:
                    listaAux.append((q*r)+((1-q)*(1-r)))
                # Que aumente o se mantenga el nivel y aumenten las horas en 1000 pero estoy en 5 así que el nivel no puede empeorar más
                elif i_==i and j_==j+1000 and i==5 and j<99500:
                    listaAux.append((q*(1-r))+((1-q)*r))
                # Que aumente o se mantenga el nivel y aumenten las horas en 500 o 1000 y llegue a 100000
                elif i_==i and j_==100000 and i==5 and j>99000:
                    listaAux.append(1)
                else:
                    listaAux.append(0)
            matriz.append(listaAux)
        matrices[acciones[0]] = np.array(matriz)
    # Para a == 1 ~ hacer mantenimiento
    elif a == 1:
        matriz = []
        for estado_actual in range(0,len(estados)):
            i = int(estados[estado_actual].split(",")[0])
            j = int(estados[estado_actual].split(",")[1])
            listaAux = []
            for estado_futuro in range(0,len(estados)):
                i_ = int(estados[estado_futuro].split(",")[0])
                j_ = int(estados[estado_futuro].split(",")[1])
                # Que se acumulen 500 horas 
                if i_==i and j_==j+500 and j<99500:
                    listaAux.append(1-r)
                # Que se acumulen 1000 horas
                elif i_==i and j_==j+1000 and j<99500:
                    listaAux.append(r)
                # Que se acumulen o 500 o 1000 horas y llegue a 100000
                elif i_==i and j_==100000 and j>99000:
                    listaAux.append(1)
                else:
                    listaAux.append(0)
            matriz.append(listaAux)
        matrices[acciones[1]] = np.array(matriz)
    # Para a == 2 ~ reemplazar
    elif a == 2:
        matriz=[]
        for estado_actual in range(0,len(estados)):
            i = int(estados[estado_actual].split(",")[0])
            j = int(estados[estado_actual].split(",")[1])
            listaAux = []
            for estado_futuro in range(0,len(estados)):
                i_ = int(estados[estado_futuro].split(",")[0])
                j_ = int(estados[estado_futuro].split(",")[1])
                # Siempre que reemplazo llego a perfecto y 0 horas
                if i_==1 and j_==0:
                    listaAux.append(1)
                else:
                    listaAux.append(0)
            matriz.append(listaAux)
        matrices[acciones[2]] = np.array(matriz)
        
# Creo el problema como un mdp
mdpBateria = dtmdp(estados, acciones, matrices, retornos, discount_factor)
print(mdpBateria.solve(0, minimize=True, method = "value_iteration"))
print(mdpBateria.solve(0, minimize=True, method="policy_iteration"))


import numpy as np
from scipy import linalg
from scipy import sparse

class dtph():

    # number of phases
    n_phases:int=1

    # substochastic matrix as 2d numpy array
    T:np.array

    # initial phase vector as 1d numpy array
    alpha:np.array
    
    # empty initializer 
    # chain with a single state
    def __init__(self):
        self.n_phases = 1
        self.T = np.array([0.8])
        self.alpha = np.array([1])

    # initializer with only the number of states
    def __init__(self,n:int):
        self.n_phases = n
        self.generator = np.zeros((n,n), dtype=float)
        self.alpha = np.zeros(n, dtype=float)
        self.alpha[0] = 1

    # initializer with both alpha and T
    def __init__(self, alpha:np.array, T:np.array):
        if not self._check_sub_stochastic_matrix(T): #Lets check if transition matrix is logical (i.e the rows sum <=1)
            raise ValueError("the rows of transition matrix do not sum <= 1")
        self.n_phases=T.shape[0]
        self.T = T
        shape = alpha.shape
        if shape[0]!=self.n_phases:
            raise ValueError("The dimensions of alpha vector are incorrect. Its length must coincide with the size of T")
        self.alpha = alpha

    def _check_sub_stochastic_matrix(self, M:np.ndarray):
        #Check if a given rate matrix has the condition that every row sum <= 0, with at least one < 0 
        if np.max(np.sum(M, axis = 1)) <= 1 and np.min(np.sum(M, axis = 1)) < 1:
            return True
        else:
            return False

    def pmf(self,n:int) -> np.float64:
        P=np.linalg.matrix_power(self.T,n)
        probs=self.alpha@P
        return sum(probs)


    def expected_value(self):
        # Compute the expected value as alpha*inv(I - T)*one
        return np.sum(self.alpha@(np.eye(self.n_phases) -  np.linalg.inv(-self.T)))

